import rasterio
import argparse
import os
import datetime
from os.path import join
import traceback

import numpy.ma as ma
import numpy as np

from frame_s2.config.FRAME_settings import lv1_settings
from frame_s2.config.FRAME_probav_settings import proba_settings
from frame_s2.postproc_helper import  tif_update_metadata, dekadToStartDate, validate_inputs, previousDekad, Spectral_file, run_command

def parse_arguments():
    parser = argparse.ArgumentParser(prog='postprocess_lv1.py')

    parser.add_argument("--mask", type=str, help="Land/Sea MASK-IMG (byte, no extension, 0=Water, 1-255=Land",
                        required=True)
    parser.add_argument("--dekad", type=str, help="dekad",
                         required=True)
    parser.add_argument("--input_dir", type=str, help="input dir for albedo processing",
                        required=True)
    parser.add_argument("--output_dir", type=str, help="input dir for albedo processing",
                        required=True)
    parser.add_argument("--normalisation", type=str, help="Divide SUM[Wi x Ri] by SUM[Wi]  [0/1=def]",
                        default=1)
    parser.add_argument("--sensor", type=str, help="sensor=MODIS(default) | PROBA",
                        default='MODIS')
    return parser.parse_args()

def calc_spectral_settings_4_albedo(bands, weights, NORM=1, sensor='MODIS'):
    VHI = 250
    Nv = len(bands)
    spectral_in_settings_lst=[]


    # Convert    Reflectances    to    fractions[0.0 - 1.0]
    for band in bands:
        for v in range(Nv):
            if sensor.upper() == 'MODIS':
                scale = lv1_settings["modis_bands"][band]["scale"]
                offset = lv1_settings["modis_bands"][band]["offset"]
                bname = lv1_settings["modis_bands"][band]["name"]
            elif sensor.upper() == 'PROBA':
                scale = proba_settings["proba_bands"][band]["scale"]
                offset = proba_settings["proba_bands"][band]["offset"]
                bname = proba_settings["proba_bands"][band]["name"]
            while (offset + scale*VHI > 2.0):
                scale = scale/10
                offset = offset/10
        spectral_in_settings_lst.append(Spectral_file(Vint=offset, Vslo=scale, Vhi=VHI, name=bname))

    #calculate sum of weights
    sum = 0.0
    for v in range(Nv):
        weight = weights[v] #put this in description
        sum+=weight
        # also take weight into account
        spectral_in_settings_lst[v].Vslo = spectral_in_settings_lst[v].Vslo * weight
        spectral_in_settings_lst[v].Vint = spectral_in_settings_lst[v].Vint * weight

    if NORM == 0: #Divide SUM[Wi x Ri] by SUM[Wi]  [0/1=def]
        sum = 1.0

    for v in range(Nv):

        spectral_in_settings_lst[v].Vslo = spectral_in_settings_lst[v].Vslo / sum
        spectral_in_settings_lst[v].Vint = spectral_in_settings_lst[v].Vint / sum

    return spectral_in_settings_lst


def calc_and_mask_albedo(in_dir, out_dir, mask_file, dekad, NORM = 1, sensor='MODIS'):
    if sensor.upper() == 'MODIS':
        BANDS = ["2", "3", "b3", "b4", "b6", "b7"]
        WEIGHTS = [0.215, 0.215, 0.242, 0.180, 0.112, 0.036]
        filename_prefix = 'mt'
        level=1

    elif sensor.upper()[0:5] == 'PROBA':
        BANDS = ["1", "2", "3", "4"]
        WEIGHTS = [0.429, 0.333, 0.133, 0.105]
        filename_prefix = 'pt'
        level=2

    albedo_file = join(out_dir, 'L%s_ALB_%s.tif') % (level, dekad)
    Nv = len(BANDS)

    spectral_in_settings_lst = calc_spectral_settings_4_albedo(BANDS, WEIGHTS, NORM=NORM, sensor=sensor)

    # Creating profile with sos file
    fname = join(in_dir, '%s%s%s.tif') % (filename_prefix, dekad, BANDS[0])
    with rasterio.open(fname, 'r', sharing=False) as src:
        profile = src.profile

    # set output
    profile.update(
        dtype=rasterio.uint8,
        count=1,
        compress='lzw',
        BIGTIFF='YES')

    print("Calculating ALBEDO ...")
    with rasterio.open(albedo_file, 'w+', **profile, sharing=False) as dst:
        with rasterio.open(mask_file, 'r', sharing=False) as msk:
            for ji, window in dst.block_windows(1):
                #print ('ji : %s, %s' % (ji[0], ji[1]))
                in_data_cube = np.full((Nv, window.height, window.width), 0, np.ubyte)
                mask_array = msk.read(1, window=window).astype(np.ubyte)
                for v in range(Nv):
                    fname = join(in_dir, '%s%s%s.tif') % (filename_prefix, dekad, BANDS[v])
                    src = rasterio.open(fname, "r", sharing=False)
                    in_array = src.read(1, window=window).astype(np.ubyte)
                    src.close()
                    in_data_cube[v, ...] = in_array
                    in_array = None
                albedo = calc_albedo_by_matrix(in_data_cube, mask_array, WEIGHTS, spectral_in_settings_lst)
                dst.write(albedo, 1, window=window)

    #update tif metadata
    albedo_file = tif_update_metadata(albedo_file,
                                   scale=lv1_settings['ALB_SCALE'],
                                   offset=lv1_settings['ALB_OFFSET'],
                                   description=lv1_settings['ALB_DESCRIPTION'],
                                   units=lv1_settings['ALB_UNITS'],
                                   flags=lv1_settings['ALB_FLAGS'],
                                   nodata=lv1_settings['ALB_NODATA']
                                    )
    print("  Finished successfully with writing %s" % albedo_file)

def calc_albedo_by_matrix(in_data_cube, ndvi_mask_array, weights, spectral_in_settings_lst):
    Nv = len(weights)

    masked = ma.masked_greater(in_data_cube, 250).astype(np.float32)
    #convert to physical values
    for i in range(Nv):
        masked[i, ...] = masked[i, ...] * spectral_in_settings_lst[i].Vslo + spectral_in_settings_lst[i].Vint
    masked_sum = masked.sum(0) #create sum over axis=0 => Physical Yalb, supposed to be in range 0.0 - 1.0
    albedo_dig = (masked_sum - lv1_settings['ALB_OFFSET']) / lv1_settings['ALB_SCALE'] # Digital  Valb
    albedo_dig = ma.floor(0.5 + albedo_dig)                                          # Valb, now rounded to INTEGER
    albedo_dig = ma.where(albedo_dig < 0, 0, albedo_dig)                               #keep track of saturation
    albedo_dig = ma.where(albedo_dig > 200, 200, albedo_dig)                          #keep track of saturation
    albedo_dig = albedo_dig.astype(np.ubyte)

    #fix masked values, if all masked => water, if not all masked then it is just missing
    count = ma.count_masked(masked, 0)
    albedo_dig = ma.filled(albedo_dig, 252)   #this should be a temp value
    albedo_dig = np.where(count > 0, 251, albedo_dig)   #if one is missing, replace the 252 value
    albedo_dig = np.where(count == Nv, 254, albedo_dig) #if all are missing,repalce the 252 value

    #apply ndvi masker
    albedo_dig = np.where(ndvi_mask_array > 250, ndvi_mask_array, albedo_dig)
    return albedo_dig


def main():
    args = parse_arguments()
    calc_and_mask_albedo(args.input_dir, args.output_dir, args.mask, args.dekad, NORM=args.normalisation, sensor=args.sensor)

if __name__ == "__main__":
    main()