# -*- coding: utf-8 -*-

import argparse
import rasterio
import numpy as np
import numpy.ma as ma
import os
from os.path import join


from frame_s2.config.FRAME_settings import meteo_settings_dict
from frame_s2.postproc_helper import  tif_update_metadata, create_quicklook, check_chronological_order_input_lst


VALIDATE_INPUT_LST=False
OUT_SCALE=0.004
OUT_OFFSET=-0.08
OUT_UNITS='-'

def parse_arguments():
    parser = argparse.ArgumentParser(prog='nok.py',description="Create IMGs with MAX. DMP/NPP from LoRes METEO-DATA")

    parser.add_argument("--in_lst", type=str, help="a file with all input file paths",
                        required=True)
    parser.add_argument("--out_lta_folder", type=str, help="output lta folder",
                        required=True)
    parser.add_argument("--roi", type=str, help="roi=AFR|SRI|SSB|ALG",
                        required=True)
    parser.add_argument("--out_file_naming", type=str, help="e.g: L2_SRI_NDVI_LTA_%dek% without extention, %dek% will be replaced by the dek of the input files",
                        required=True)

    return parser.parse_args()

def calc_lta_by_matrix(data_cube, in_min=0, in_max=250, out_nodata=255):
    #the easy way is using np.mean, but we cannot filter good enough
    n, rows, columns = data_cube.shape
    sum_array = np.full((rows, columns), 0)
    count_array = np.full((rows, columns), 0)

    masked = ma.masked_greater(data_cube, 250)
    mean = ma.mean(masked, 0)
    out_array = np.floor(mean + 0.5).astype(np.uint8)  # use rounding half up method
    out_array = out_array.filled(fill_value=out_nodata)
    return out_array
def validate_inputs(in_img_list):
    bbox_list=[]
    shape_list=[]
    for img in in_img_list:
        with rasterio.open(img) as src:
            shape_list.append(src.shape)
            bbox_list.append(src.bounds)

            #validating blocksize
            blockxsize = int(src.profile["blockxsize"])
            blockysize = int(src.profile["blockysize"])
            if blockxsize != meteo_settings_dict["input_blockXsize"] or blockysize != meteo_settings_dict["input_blockYsize"]:
                raise Exception ("Input blocksize %s,%s does not match the expected ones %s,%s"
                                 % (blockxsize,
                                    blockysize,
                                    meteo_settings_dict["input_blockXsize"],
                                    meteo_settings_dict["input_blockYsize"]))

    #validate if all items in the list are the same, by using set
    shape_set = set(shape_list)
    bbox_set = set(bbox_list)
    if len(shape_set)!= 1: raise Exception ('Shapes of the in images has differences : %s for files %s' % (shape_list, in_img_list))

    #check if all coordinates of the bbox are equal, keep in mind that some rounding errors can be there
    left=[];right=[];top=[];bottom=[]
    digits=7    #7 digits is less than a meter (for 100m data)
    for bbox in bbox_list:
        left.append(round(bbox[0], digits))
        bottom.append(round(bbox[1], digits))
        right.append(round(bbox[2], digits))
        top.append(round(bbox[3], digits))
    left_set=set(left)
    bottom_set=set(bottom)
    right_set=set(right)
    top_set=set(top)

    #if len(bbox_set) != 1: raise Exception('Shapes of the in images has differences : %s' % bbox_list)
    if len(left_set) != 1 or len(bottom_set)!= 1 or len(right_set)!=1 or len(top_set)!=1 :
        raise Exception('Shapes of the in images has differences : %s' % bbox_list)


    total_number_of_blocks = round(int(shape_list[0][0])/blockxsize * int(shape_list[0][1])/blockysize)
    if total_number_of_blocks == 1: print("-Warn- Total number of blocks is 1! For big images this can couse memory" 
                                               " issues. To solve this, provide is tiff with blocks of eg 256:256")
    return total_number_of_blocks

def main():
    args = parse_arguments()
    start_lta(args.in_lst, args.out_lta_folder,  args.out_file_naming , out_scale=OUT_SCALE, out_offset=OUT_OFFSET, out_units=OUT_UNITS)

def start_lta(in_lst, out_folder, out_file_suffix, out_scale=1, out_offset=0, out_units="-"):
    #set gdal cache parameter on the workers to reduce memory usage.
    # rasterio uses gdal in the background so limit the GDAL_CACHE
    # this parameter is in bytes. 1GB wil result in 1,2GB of memory usage(tested on local machine)
    os.environ['GDAL_CACHEMAX'] = meteo_settings_dict['GDAL_CACHEMAX']
    colormap="ndvi.txt"
    ql_scale = None
    if out_file_suffix.find('SRI')>0:
        ql_width= 0
        ql_height = 1000

    else:
        ql_width= 0
        ql_height= 1000

    # handling IN's
    check_chronological_order_input_lst(in_lst)
    # create a list of NDVI input files
    with open(in_lst, 'r') as f:
        in_file_lst = f.readlines()
    in_file_lst = list(map(lambda ndvi: ndvi.strip(), in_file_lst))  # remove line breaks from list
    #in_file_lst.sort() #NEVER EVER SORT this list!!!!

    print("Number of input files : %s" % len(in_file_lst))


    #Parse filenames and devide by dekad, filename should ALWAYS end with YYTT.tif
    dek_dict={}
    for i in range(1, 37):
        dek_dict[str(i).zfill(2)]=[]
    for in_file in (in_file_lst):
        fname = os.path.basename(in_file)
        yytt = fname[-8:-4]
        if not yytt.isdigit():
            raise Exception("-E- Unable to get the dekad (%s) from the filename %s" % (yytt, fname))
        yy=yytt[0:2]
        tt=yytt[2:4]
        print (tt)
        dekad_files = dek_dict[tt]
        dekad_files.append(in_file)
        dek_dict[tt] = dekad_files

    #investigate dict. All dekads needs the same amounf of file
    #loop over dekads, and count infiles
    if VALIDATE_INPUT_LST:
        nr_01=len(dek_dict['01'])
        for i in range(2,37):
            i=str(i).zfill(2)
            nr_i = len(dek_dict[i])
            if nr_i != nr_01:
                raise Exception("There is a difference in lenght for the dekads. Nr of dekads for dekad '01'=%s; Nr of dekads for dekad '%s'=%s" % (nr_01,i, nr_i))

    print("Start processing...")

    # validation
    total_number_of_blocks = validate_inputs(in_file_lst)



    # Creating profile with the first image from the list
    with rasterio.open(in_file_lst[0], 'r', sharing=False) as src_sr:
        sr_profile = src_sr.profile

        # set output
        profile = sr_profile
        profile.update(
            dtype=rasterio.uint8,
            count=1,
            compress='lzw',
            BIGTIFF='YES')

    for dekad, dekad_files in dek_dict.items():
        in_fname_lst=list(map(lambda full_path: os.path.basename(full_path), dekad_files))
        lst_lenght=len(dekad_files)
        if lst_lenght==0:
            #skip empty dekads
            continue
        fname=out_file_suffix.replace("<dek>", dekad) + '.tif'
        out_fname = os.path.join(out_folder, fname)
        print (out_fname)
        # Creating datacube_in with all in files per tif block
        with rasterio.open(out_fname, 'w',**profile, sharing=False) as dst:
            for ji, window in dst.block_windows(1):
                in_data_cube = np.full((len(dekad_files), window.height, window.width), 0, np.uint8)

                # reading all inputs and fill up in_data_cube
                for index, NDVI_file in enumerate(dekad_files):
                    # print ("Handling infile %s" % NDVI_file)
                    src = rasterio.open(NDVI_file, 'r', sharing=False)
                    in_array = src.read(1, window=window).astype(np.uint8)
                    src.close()
                    in_data_cube[index, ...] = in_array
                    in_array = None

                # Start processing
                lta_array = calc_lta_by_matrix(in_data_cube)
                #lta_array = np.mean(data_cube, 0 )
                data_cube = None

                dst.write(lta_array, 1, window=window)

            # #update out metadata
            offset = out_offset
            scale = out_scale
            units = out_units
            description = "Long-Term Average, computed for Dekad=%s, over %s tifs : %s" % (dekad, lst_lenght, in_fname_lst)

        out_fname = tif_update_metadata(out_fname, scale=(scale or 1), offset=(offset or 0), description=description, units=units)
        create_quicklook(out_fname, colormap, height=ql_height, width=ql_width, scale=ql_scale)
        print("Finished successfully with writing %s" % out_folder)

if __name__ == "__main__":
    main()