# -*- coding: utf-8 -*-

import argparse
import rasterio
import numpy as np
import os
from ctypes import c_double, c_short, c_float, CDLL, Structure, c_ubyte

from frame_s2.config.FRAME_settings import meteo_settings_dict, lv2_settings
from frame_s2.postproc_helper import validate_inputs, Spectral_file, tif_update_metadata

so_file='FREQ_EVENT_MEP2.so'
if not os.path.exists(so_file):
    so_file = 'executables/FREQ_EVENT_MEP2.so'
lib = CDLL(so_file)
lib_loop = lib.loop  # c_sum is the name of our C function
lib_loop.argtypes = [
    np.ctypeslib.ndpointer(dtype=np.ubyte, ndim=2, flags='C_CONTIGUOUS'),
    np.ctypeslib.ndpointer(dtype=np.single, ndim=3, flags='C_CONTIGUOUS'),
    np.ctypeslib.ndpointer(dtype=np.short, ndim=2, flags='C_CONTIGUOUS'),
    np.ctypeslib.ndpointer(dtype=np.single, ndim=1, flags='C_CONTIGUOUS'),
    c_short, c_short, c_short, Structure]
lib_loop.restype = None

lib_version = lib.print_version
lib_version.restype = None
lib_version.argtypes = None

class HdrInfo(Structure):
    _fields_ = [
        ("Vlo", c_double),
        ("Vhi", c_double),
        ("Vmax", c_double),
        ("Vmin", c_double)
    ]


class NokSettings(Structure):
    _fields_ = [
        ("MskVal", c_ubyte),
        ("outside", c_short),
        ("TYPo", c_short),
        ("DTo", c_short),
        ("OutHI", c_short),
        ("flagM", c_short),
        ("flagO", c_short),
        ("InLO", c_float),
        ("InHI", c_float),
        ("EvtLO", c_float),
        ("EvtHI", c_float),
        ("factor", c_float)
    ]


def parse_arguments():
    parser = argparse.ArgumentParser(prog='nok.py',description="Create IMGs with MAX. DMP/NPP from LoRes METEO-DATA")

    parser.add_argument("--in_file_lst", type=str, help="a file with all NDVI/QUA paths",
                        required=True)
    parser.add_argument("--out_nok", type=str, help="output nok file",
                        required=True)
    parser.add_argument("--out_datatype", type=int,
                        help="0 : auto(depending on nr of inputs), 1: byte, 2: integer",
                        required=False,
                        default=0)
    parser.add_argument("--in_mask", type=str, help="OPTIONAL MASK IMG (BYTE: 0=MASK-OUT, 1-255=KEEP-IN)",
                        required=False,
                        default=None)
    parser.add_argument("--out_flag_masked_pixels", type=int, help="OUT-Flag for Masked pixels",
                        required=False,
                        default=255)

    parser.add_argument("--in_Vlo", type=float, help="lowest non-flag value for input images, default=0",
                        required=False,
                        default=0)
    parser.add_argument("--in_Vhi", type=float, help="highest non-flag value for input images, default=250",
                        required=False,
                        default=250)
    parser.add_argument("--out_flag_all_flagged_in", type=int, help="OUT-Flag for All-Flagged Pixels",
                        required=False,
                        default=0)

    parser.add_argument("--calculation_type", type=int, help="0=Abs, 1=Rel.to non-flagged, 2=rel. to total obs. (default)",
                        required=False,
                        default=2)
    parser.add_argument("--count_ins", type=int,
                        help="Count if IMG-value is IN range LO-HI (0=default) or BEYOND this range (1)",
                        required=False,
                        default=0)
    parser.add_argument("--lo_val", type=float, help="DIGITAL RANGE OF VALUES TO CONSIDER AS 'EVENT' IN THE FREQUENCY COUNTING, LO_val",
                        required=False,
                        default=0)
    parser.add_argument("--hi_val", type=float, help="DIGITAL RANGE OF VALUES TO CONSIDER AS 'EVENT' IN THE FREQUENCY COUNTING, HI_val",
                        required=False,
                        default=250)
    return parser.parse_args()


def calc_nok_by_matrix(data_cube, mask_array, in_Vlo, in_Vhi, out_flag_masked_pixels, out_flag_all_flagged_in, event_lo_val, event_hi_val, out_spectral_file, count_ins, factor=100.0, calculation_type=0 ):
    #calculate nok with matrix Nv x rows x columns
    if out_spectral_file.datatype == np.uint8:
        DTo=1
    else:
        DTo=2

    data_cube = data_cube.astype(np.single)     # data_cube in float
    n, rows, columns = data_cube.shape
    Settings = NokSettings(1, count_ins, calculation_type, DTo, out_spectral_file.Vhi, out_flag_masked_pixels, out_flag_all_flagged_in, in_Vlo, in_Vhi, event_lo_val, event_hi_val, factor)

    #initialize C-arrays
    result_cube = np.empty([rows, columns], np.short)
    arr_pix = np.empty(n, np.single)

    # start C-processing
    lib_loop(mask_array, data_cube, result_cube, arr_pix, n, rows, columns, Settings)
    return result_cube


def main():
    args = parse_arguments()
    start_nok(args.in_file_lst, args.in_mask, args.in_Vlo, args.in_Vhi, args.out_flag_masked_pixels,
              args.out_flag_all_flagged_in, args.out_nok, args.calculation_type, args.lo_val,
              args.hi_val, args.out_datatype, args.count_ins)

def start_nok(in_file_lst_file, in_mask, in_Vlo, in_Vhi, out_flag_masked_pixels, out_flag_all_flagged_in, out_nok, calculation_type, lo_val, hi_val, out_datatype, count_ins):
    #set gdal cache parameter on the workers to reduce memory usage.
    # rasterio uses gdal in the background so limit the GDAL_CACHE
    # this parameter is in bytes. 1GB wil result in 1,2GB of memory usage(tested on local machine)
    os.environ['GDAL_CACHEMAX'] = meteo_settings_dict['GDAL_CACHEMAX']


    # data_cube=np.array([
    #         [[0, 20],
    #          [30, 150]
    #      ],[
    #         [50, 60],
    #         [70, 255]]])


    #for AFR, the roi abbreviation is not in the filenames, for the other roi's it is.
    # if roi == 'AFR':
    #     roi_prefix = 'L2'
    # else:
    #     roi_prefix = 'L2_%s' % roi


    #create a list of NDVI input files
    with open(in_file_lst_file, 'r') as f:
        in_file_lst = f.readlines()
    in_file_lst = list(map(lambda in_file: in_file.strip(), in_file_lst)) #remove line breaks from list
    print ("Number of input files : %s" %len(in_file_lst))
    print ("Mask file : %s " % in_mask)
    print ("Out file  : %s " % out_nok )
    print ("Start processing...")
    lib_version() #printing version to out

    #set spectral parameters for out file
    if count_ins:
        inside_txt = "inside"
    else :
        inside_txt = "outside"
    flags = "%s:all flagged IN, %s:masked using masker %s" %(out_flag_all_flagged_in, out_flag_masked_pixels, os.path.basename(in_mask))

    out = Spectral_file(Vlo=0, Vhi=200, Vint=0, Vslo=0.5, datatype=rasterio.uint8, flags=flags)  #inital settings will be changed in next lines
    if out_datatype == 0:   #auto detect, depending on nr of inputs use uint8 or int16
        if len(in_file_lst) > 250:
            out.Vlo = 0; out.Vhi = 10000; out.Vint = 0; out.Vslo = 0.01; out.datatype=rasterio.int16
    elif out_datatype == 2:
        out.Vlo = 0; out.Vhi = 10000; out.Vint = 0; out.Vslo = 0.01;out.datatype=rasterio.int16

    if calculation_type == 0 : #absolute frequency
        out.Vlo = 0; out.Vhi = len(in_file_lst); out.Vint = 0; out.Vslo = 1; out.units = '-';
        out.description = "Counted Events if digital value %s range [%s ... %s]" % (inside_txt, lo_val, hi_val)
    elif calculation_type == 1: #Relative frequencies in %
        out.description = "Fraction (Events / Non-flagged IN-values) if digital value %s range [%s ... %s]" % (inside_txt, lo_val, hi_val)
    elif calculation_type == 2: #"Fraction (Events / Total IN-values)"
        out.description = "Fraction (Events / Total IN-values) if digital value %s range [%s ... %s]" % (inside_txt, lo_val, hi_val)

    factor = 100 / out.Vslo


    #validation of inputfiles & mask : spatial and spectral
    total_number_of_blocks = validate_inputs(in_file_lst)
    total_number_of_blocks = validate_inputs( [in_file_lst[0], in_mask], spatial_validation=True, spectral_validation=False )   # see if mask and inputs are maching spatially

    # Creating profile with the first image from the list
    with rasterio.open(in_file_lst[0], 'r', sharing=False) as src_sr:
        sr_profile = src_sr.profile

        #Update output profile
        profile = sr_profile
        profile.update(
            dtype=out.datatype,
            count=1,
            compress='lzw',
            BIGTIFF='YES')

    #Start opening all files
    if in_mask is not None:
        msk = rasterio.open(in_mask, 'r', sharing=False)
    with rasterio.open(out_nok, 'w', **profile, sharing=False) as dst:
        for ji, window in dst.block_windows(1):
            print('ji : %s, %s' % (ji[0], ji[1]))
            data_cube = np.full( (len(in_file_lst), window.height, window.width), out_flag_masked_pixels, np.single)    #init cube with masked values

            # opening mask if it is there
            mask_array = np.full((window.width, window.height), 1, np.uint8) #by default values are 1 (not masked)
            if in_mask is not None:
                mask_array = msk.read(1, window=window).astype(np.uint8)

            #open all input files and put there blocks into the datacube.
            for index, in_file in enumerate(in_file_lst, start=0):
                src = rasterio.open(in_file, 'r', sharing=False)
                in_array = src.read(1, window=window).astype(np.uint8)
                src.close()
                data_cube[index, ...] = in_array
                in_array = None


            #testarray
            # data_cube = np.full((3, 2, 2), 255)
            # mask_array = np.array([[0, 1], [1, 1]],np.uint8)
            # data_cube[0, ...] = np.array([[4, 5], [6, 7]], np.single )
            # data_cube[1, ...] = np.array([[0, 255], [30, 150]])
            # data_cube[2, ...] = np.array([[50, 60], [70, 255]])

            # processing

            nok = calc_nok_by_matrix(data_cube, mask_array, in_Vlo, in_Vhi, out_flag_masked_pixels, out_flag_all_flagged_in, lo_val, hi_val, out, count_ins, factor=factor, calculation_type=calculation_type)
            mask_array = None
            data_cube = None

            dst.write(nok.astype(out.datatype), 1, window=window)

        dst.nodata = None
        msk.close()

    #set tif metadata
    out_nok = tif_update_metadata(out_nok, scale=(out.Vslo or 1), offset=(out.Vint or 0), description=out.description,
                                   units=out.units, flags=out.flags)

    print("Finished successfully with writing %s" % out_nok)

if __name__ == "__main__":
    main()