# -*- coding: utf-8 -*-

import argparse
import rasterio
import numpy as np
import os, sys
from os.path import join
from math import floor, ceil
from ctypes import c_double, c_short, c_float, CDLL, Structure, c_ubyte,POINTER, c_char, c_long, c_wchar_p, c_char_p

from frame_s2.config.FRAME_settings import lv1_settings
from frame_s2.postproc_helper import validate_inputs, getHistoricalDekadList, Spectral_file, tif_update_metadata, \
    parse_date, dateToDekad, previousDekad, nextDekad, dekadToStartDate, get_days_in_dekad, D50, getDekadLst

class HdrInfo(Structure):
    _fields_ = [
        ("Vlo", c_double),
        ("Vhi", c_double),
        ("Vint", c_double),
        ("Vslo", c_double)
    ]

class HdrInfo_in(Structure):
    _fields_ = [
        ("fi", c_wchar_p), #c_char_p
        ("offset", c_long),
        ("date", c_long),
        ("ok", c_short),
        ("D50", c_double)
    ]

class HdrInfo_out(Structure):
    _fields_ = [
        ("fo", c_wchar_p), #c_char_p
        ("fq", c_wchar_p),
        ("date", c_long)
    ]


class SmoothSettings(Structure):
    _fields_ = [
        ("METsmooth",c_short),
        ("METflag", c_short),
        ("Extrapol", c_short),
        ("NviMaxMis", c_short),
        ("QUA", c_short),
        ("YmaxDip", c_float),
        ("Ydelt", c_float),
        ("D50pre", c_double),
        ("D50nxt", c_double),

        ("PreMaxDip", c_float),
        ("PreMaxDif", c_float),
        ("PreMaxTop", c_float),
        ("PreMaxGap", c_short),
        ("PreMaxGapApply", c_short),
        ("SnowMsk", c_short),


        ("MUinterpol", c_short),
        ("MUrmf", c_short),
        ("SwWmax", c_float),
        ("SwWmin", c_float),
        ("SwWplane", c_float),
        ("SwWslope", c_float),
        ("SwWedge", c_float),
        ("SwCI", c_float),
        ("SwWinR", c_short),
        ("SwWinC", c_short),
        ("WinRl", c_short),
        ("WinRr", c_short),
        ("WinCl", c_short),
        ("WinCr", c_short),

        ("Post", c_float),
        ("PostMax", c_float),
        ("PostOver", c_short),
        ("PostUnder", c_short),

        ("Ymin", c_float),
        ("VIflag", c_float),
        ("FLAGi", c_ubyte),
        ("vio1", c_short),

        ("PreMaxGapMsk", c_char_p ),  # c_char_p van 260
        ("Ny", POINTER(c_short)),
        ("Yrm", POINTER(c_float)),
        ("SW", POINTER(c_float)),
        ("Yw", POINTER(c_float)),
        ("CHI2T", POINTER(c_float) * 201)
    ]



#hack for running smooth15.so on cluster and on jenkins
so_file='smooth15.so'
if not os.path.exists(so_file):
    so_file = 'executables/smooth15.so'

lib = CDLL(so_file)
lib_loop = lib.loop  # c_sum is the name of our C function
lib_loop.argtypes = [
    #in
    np.ctypeslib.ndpointer(dtype=np.ubyte, ndim=3, flags='C_CONTIGUOUS'), c_short,
    #out
    np.ctypeslib.ndpointer(dtype=np.ubyte, ndim=3, flags='C_CONTIGUOUS'),
    np.ctypeslib.ndpointer(dtype=np.ubyte, ndim=3, flags='C_CONTIGUOUS'), c_short,
    np.ctypeslib.ndpointer(dtype=np.short, ndim=1, flags='C_CONTIGUOUS'),
    #set
    c_short, c_short]
lib_loop.restype = None

#settings the global Settings structure, these settings are hard coded in C
lib_Allocate_mem = lib.Allocate_mem
lib_Allocate_mem.argtypes = [c_short, c_short]
lib_Allocate_mem.restype = None

lib_settings = lib.setSettings
lib_settings.argtypes = [HdrInfo, HdrInfo, SmoothSettings, POINTER(HdrInfo_in), POINTER(HdrInfo_out) ]
lib_settings.restype = None

# lib_HI = lib.Set_inHeaders
# lib_HI.argtypes=[HdrInfo_in, c_short]
# lib_HI.restype = None


def parse_arguments():
    parser = argparse.ArgumentParser(prog='smooth15.py',description="Smoothing MODIS & PROBA: smoothing of NDV (Swets) and ALB (MEAN)")
    #IN-settings
    parser.add_argument("--ndvi", type=int, help="NDVI (1) or ALBEDO (0)",
                        required=False,
                        default=1)
    parser.add_argument("--in_file_prefix", type=str, help="dir + file prefix",
                        required=True)
    parser.add_argument("--in_file_suffix", type=str, help="file suffix (i for ndvi,...",
                        required=False,
                        default='')
    parser.add_argument("--start_in", type=str, help="Start IN-series YYYYMMDD",
                        required=True)
    parser.add_argument("--end_in", type=str, help="End IN-series YYYYMMDD",
                        required=True)
    parser.add_argument("--max_missing", type=int, help="Max. Nr. of consec. missing IN-IMGs, unit=%",
                        required=False,
                        default=0)
    parser.add_argument("--hdr_in", type=str, help="scale offset vlo vhi",
                        required=False,
                        default="1 0 0 250")

    #OUT settings
    parser.add_argument("--out_file_prefix", type=str, help="outputdir + file prefix for smoothed data",
                        required=True)
    parser.add_argument("--out_file_suffix", type=str, help="k for NDVI?",
                        required=False,
                        default='')
    parser.add_argument("--start_out", type=str, help="Start OUT-series YYYYMMDD",
                        required=True)
    parser.add_argument("--end_out", type=str, help="End OUT-series YYYYMMDD",
                        required=True)


    #OTHER PARAMETERS concerning the TIME SERIES OF IN/OUT-IMGs
    parser.add_argument("--ymin", type=float, help="Lowest physical Y-value for clear land      [def=0.0]",
                        required=False,
                        default=0.0)
    parser.add_argument("--ymaxdip", type=float, help="YmaxDip for elimination of local minima  [def=0=skip]",
                        required=False,
                        default=0)
    parser.add_argument("--nvimaxmix", type=int, help="Max. % of missing values per pixel-profile   [def=75]",
                        required=False,
                        default=75)
    parser.add_argument("--out_qua_file_prefix", type=str, help="outputdir + file prefix for smoothed qua data. Leave empty or 0 to disable QUA-images",
                        required=False,
                        default='0')
    parser.add_argument("--ydelt", type=float, help="Max. difference Ydelt       [def=0.0]",
                        required=False,
                        default=0.0)
    return parser.parse_args()

def calc_smooth_by_matrix(data_cube, NvO, stats ):
    #initialize C-arrays
    NvI, rows, columns = data_cube.shape
    result_cube = np.empty((NvO, rows, columns), np.uint8)
    qua_cube = np.empty((NvO, rows, columns), np.uint8)

    # start C-processing
    lib_loop(data_cube, NvI, result_cube, qua_cube, NvO, stats, rows, columns)
    return result_cube, qua_cube, stats

def main():
    args = parse_arguments()
    start_smooth15(args.ndvi, args.in_file_prefix, args.in_file_suffix, args.start_in, args.end_in,
                   args.max_missing, args.hdr_in,
                   args.out_file_prefix, args.out_file_suffix, args.start_out, args.end_out, args.out_qua_file_prefix,
                   args.ymin, args.ymaxdip, args.nvimaxmix,  args.ydelt)


def start_smooth15(ndvi, in_file_prefix, in_file_suffix, start_in, end_in,
                   max_missing, hdr_in,
                   out_file_prefix, out_file_suffix, start_out, end_out,
                   out_qua_file_prefix,
                   ymin, ymaxdip, nvimaxmix , ydelt):
    #set gdal cache parameter on the workers to reduce memory usage.
    # rasterio uses gdal in the background so limit the GDAL_CACHE
    # this parameter is in bytes. 1GB wil result in 1,2GB of memory usage(tested on local machine)
    os.environ['GDAL_CACHEMAX'] = lv1_settings['GDAL_CACHEMAX']

    #Check parameters
    QUA = 0
    if out_qua_file_prefix is not None or out_qua_file_prefix != '' or int(out_qua_file_prefix)==0:
        QUA = 1
    if not os.path.exists(os.path.dirname(in_file_prefix)):
        raise Exception("Input dir does not exist %s " % os.path.dirname(in_file_prefix))
    if not os.path.exists(os.path.dirname(out_file_prefix)):
        os.makedirs(os.path.dirname(out_file_prefix))

    if start_in > end_in :
        raise Exception("There is something wrong with the IN dates: start=%s, end=%s" %(start_in, end_in))
    if start_out > end_out:
        raise Exception("There is something wrong with the OUT dates: start=%s, end=%s" %(start_out, end_out))

    in_file_prop_lst, dekad_lst = get_available_in_files(in_file_prefix, in_file_suffix, start_in, end_in)

    HI_lst, HO_lst, Settings = setup_smooth(ndvi, in_file_prop_lst, start_in, end_in, max_missing,
                   out_file_prefix, out_file_suffix, start_out, end_out,
                   out_qua_file_prefix, ymin, ymaxdip, nvimaxmix, ydelt, dekad_lst)


    NvI = len(HI_lst)
    NvO = len(HO_lst)

    #convert list to an C-array
    HI = (HdrInfo_in * len(HI_lst))(*HI_lst)
    HO = (HdrInfo_out * len(HO_lst))(*HO_lst)

    if ndvi:
        hi = HdrInfo(Vlo=0, Vhi=250, Vint=-0.08, Vslo=0.004)
        ho = HdrInfo(Vlo=20, Vhi=250, Vint=-0.08, Vslo=0.004)
    else:
        hi = HdrInfo(Vlo=0, Vhi=200, Vint=0.0, Vslo=0.005)
        ho=hi

    lib_Allocate_mem(NvI, NvO)

    lib_settings(hi, ho, Settings,  HI, HO )
    # for index, hdr_in in enumerate(HI_lst):
    #     lib_HI(hdr_in, index)


    in_file_lst=[]

    for hi in HI_lst:
        if hi.ok:
            in_file_lst.append(hi.fi)

    # Creating profile with the first image from the list
    with rasterio.open(in_file_lst[0], 'r', sharing=False) as src_sr:
        sr_profile = src_sr.profile
        total_pixels = src_sr.width* src_sr.height
        nr_blocks = ceil(src_sr.width / int(src_sr.block_shapes[0][0])) * ceil(src_sr.height / int(src_sr.block_shapes[0][1]))
        progress_step = nr_blocks/100

        # Update output profile
        profile = sr_profile
        profile.update(
            dtype=rasterio.uint8,
            count=1,
            compress='lzw',
            BIGTIFF='YES')

    # validation of inputfiles & mask : spatial and spectral TODO: enable this again!!! JIRA-GDD-1924
    #total_number_of_blocks = validate_inputs(in_file_lst)

    # Start opening out files

    dst_lst=[]; qua_lst=[]

    for ho in HO_lst:
        out_name = ho.fo;
        dst = rasterio.open(out_name, 'w+', **profile, sharing=False)
        dst_lst.append(dst);

        if QUA:
            qua_name = ho.fq
            qua = rasterio.open(qua_name, 'w+', **profile, sharing=False)
            qua_lst.append(qua)

    #setting simple hdr
    #set spectral headers for inputs

    #Creating datacube_in with all NDVI files per tif block and calculate pheno
    print ("Building datacube...")
    stats = np.empty(4, np.short)

    #prepare progressbar
    print ("0% ", end="")
    for i in range(92):
        print(".", end="")
    print(" 100%")
    step=0
    with rasterio.open(in_file_lst[0], 'r', sharing=False) as src0:
        for ji, window in src0.block_windows(1):
            #udpate progress
            step+=progress_step
            if step/nr_blocks*100 > 1:
                print(".", end="")
                step=0
            #print (ji)
            # init input data_cube
            in_data_cube = np.full( (len(in_file_lst), window.height, window.width), 0, np.uint8)

            #reading first input image
            in_array = src0.read(1, window=window).astype(np.uint8)
            in_data_cube[0, ...] = in_array

            # reading all other inputs and fill up in_data_cube
            for index, in_file in enumerate(in_file_lst[1:], start=1):
                #print ("Handling infile %s" % NDVI_file)
                src = rasterio.open(in_file, 'r', sharing=False)
                in_array = src.read(1, window=window).astype(np.uint8)
                src.close()
                in_data_cube[index, ...] = in_array
                in_array = None

            # Start processing
            result_array, qua_array, stats = calc_smooth_by_matrix(in_data_cube, NvO, stats )
            in_data_cube = None

            #writing blocks to files
            #print(dst_lst)
            for index, dst in enumerate(dst_lst):
                #print("Wrting %s" %dst.files[0])
                dst.write(result_array[index].astype('uint8'), 1, window=window)
                # writing blocks to files
            if QUA:
                for index, qua in enumerate(qua_lst):
                    qua.write(qua_array[index].astype('uint8'), 1, window=window)

        print(".")  #closing progressbar
    # closing outfile
    for index, dst in enumerate(dst_lst):
        dst.nodata = 255
        dst.close()

    if QUA:
        for index, qua in enumerate(qua_lst):
            qua.nodata = lv1_settings['QUA_NODATA']
            qua.close()

    # set tif metadata, TODO, move up on level!!!???
    #qua flags     "QUA_FLAGS": "0-240 = gap since previous observation, 250 = acceptable, >250 missing",
    # out_nok = tif_update_metadata(out_nok, scale=(out.Vslo or 1), offset=(out.Vint or 0), description=out.description,
    #                               units=out.units, flags=out.flags)

    print("Finished successfully with writing %s" % out_file_prefix)
    print ("---- STATS---- ")
    f = 100 / total_pixels
    print("Pixels: Total  : %s (%7.3f%%)                              "     %( total_pixels, f * total_pixels))
    print("        Water  : %s  (%7.3f%%)   Vflag=254 (No data at all)"     % (stats[1], f * stats[1]))
    print("        Missing: %s  (%7.3f%%)   Vflag=251 (Not enough data)"    % (stats[2], f * stats[2]))
    print("        Failed : %s (%7.3f%%)   Vflag=251 (Smooth failed) "      % (stats[3], f * stats[3]))
    print("        Normal : %s  (%7.3f%%)   V=0-250                    "    % (stats[0], f * stats[0]))

class SmoothInFile():
    def __init__(self, fname, dekad, status):
        self.fname = fname
        self.dekad = dekad
        self.status = status


def get_available_in_files(in_file_prefix, in_file_suffix, startdate, enddate):
    in_file_prop_lst=[] #
    startdekad = dateToDekad(startdate)
    enddekad = dateToDekad(enddate)
    dekad_lst = getDekadLst(startdekad, enddekad)
    NvI = len(dekad_lst)  # NvI=dekads for which ACT IN-IMGs should exist - though gaps allowed
    if NvI < 18:
        raise Exception("IN-series must at least comprise 18 dekads")


    NvA = 0 #number of Actual availabe files
    for dekad in dekad_lst:
        fname = join(in_file_prefix + dekad + in_file_suffix + '.tif')
        ok = 0  # init
        if os.path.exists(fname):
            ok = 1
            NvA += 1
        in_file_prop_lst.append(SmoothInFile(fname, dekad, ok))

    print ("Nr. of IN-IMGs Found / Missing: %s / %s" % (NvA, NvI - NvA))
    if NvA < 9:
        raise Exception("Insufficient actual IN-IMGs found (at least 9)")

    return in_file_prop_lst, dekad_lst

# def get_available_in_files(input_dir, dekad_lst, product_type, sensor_type):
#     in_lst=[]
#     nok = 0   #number of Not Ok(missing) files
#     for dekad in dekad_lst:
#         if sensor_type.upper() == 'MODIS':
#             fname = join(input_dir, 'L1_%s_%s.tif') % (product_type, dekad)
#         if sensor_type.upper() == 'PROBAV':
#             #TODO : is this the same for PROBAV???
#             fname = join(input_dir, 'L1_%s_%s.tif') % (product_type, dekad)
#
#         if os.path.exists(fname):
#             in_lst.append(fname)
#         else :
#             nok+=1
#
#     return nok, in_lst

def setup_smooth(ndvi, in_file_prop_lst, start_in, end_in, max_missing,
                 out_file_prefix, out_file_suffix, start_out, end_out,
                 out_qua_file_prefix,
                 ymin, ymaxdip, nvimaxmix, ydelt, dekad_lst):

    print ('start_out : %s' % start_out)
    print ('end_out : %s' % end_out)
    NvI = len(in_file_prop_lst)
    #Settings = SmoothSettings()
    if ndvi:
        hv = HdrInfo_in(Vlo=0, Vhi=250, Vint=-0.08, Vslo=0.004)
    else:
        hv = HdrInfo_in(Vlo=0, Vhi=200, Vint=0.0, Vslo=0.005)

    #CALCULATE INPUT PERIOD
    yyyy, yy, mm, m, tty, ttm, dd, doy, dd_st, dpm, did = parse_date(start_in)
    Idate1 = yyyy + mm + dd_st      # Reset to start of dekad!
    startdekad = dateToDekad(Idate1)
    yyyy, yy, mm, m, tty, ttm, dd, doy, dd_st, dpm, did = parse_date(end_in)
    Idate2 = yyyy + mm + dd_st      #Reset to start of dekad!
    enddekad = dateToDekad(Idate2)




    #D50 of central days of dekads before Idate1 / after Idate2
    dekad = dateToDekad(start_in)
    pre_dekad = previousDekad(dekad)
    pre_start_date = dekadToStartDate(pre_dekad)
    days_in_pre_dekad = get_days_in_dekad(pre_start_date)
    Settings_D50pre = D50(pre_start_date) + days_in_pre_dekad/2
    dekad = dateToDekad(end_in)
    nxt_dekad = nextDekad(dekad)
    nxt_start_date = dekadToStartDate(nxt_dekad)
    days_in_nxt_dekad = get_days_in_dekad(nxt_start_date)
    Settings_D50nxt = D50(nxt_start_date) + days_in_nxt_dekad / 2

    #check actual in images
    NvA = 0 #existing in-images found
    yyyy, yy, mm, m, tty, ttm, dd, doy, dd_st, dpm, did = parse_date(start_in)
    dd = int(dd) - 10 #Initialize: Reset date counters to Idate1 - 10 days
    HI=[]
    for smoothFile in in_file_prop_lst:
        dekad = smoothFile.dekad
        dekad_date = dekadToStartDate(dekad)
        yyyy, yy, mm, m, tty, ttm, dd, doy, dd_st, dpm, did = parse_date(dekad_date)
        f = get_days_in_dekad(dekad_date)
        d50 = float(D50(dekad_date) + f/2 - Settings_D50pre)     #Make all D50-dates relative to D50pre=0
        #fname = join(in_file_prefix + dekad + in_file_suffix + '.tif')
        # ok = 0 #init
        # if os.path.exists(fname):
        #     ok = 1
        #     NvA += 1
        HI.append(HdrInfo_in(smoothFile.fname, 0, int(dekad_date), smoothFile.status, d50 ))




    Settings_D50nxt = Settings_D50nxt - Settings_D50pre         #Make all D50-dates relative to D50pre=0
    Settings_D50pre = 0.0

    #check on maxGap
    NvG = 0                 #NvG=Total lacking IMGs in centre (those with ok=0), gap=longest gap found
    gap = 0
    i = HI[0].ok            #i=previous IMG absent/present, k=number of IMGs in current gap
    k = 0
    for v in range(NvI):
        j = HI[v].ok        #j = type of current IMG (1=OK, 0=lacking)
        if i == 0 & j == 0:
            k += 1
            NvG += 1
        if i == 1 & j == 0:
            k = 1
            NvG += 1
        if i == 0 & j == 1:
            if k > gap:
                gap = k
                k = 0
    if k > gap:
        gap = k
        k = 0
    print ("Widest gap : %d dekads" % gap)
    if gap > max_missing:
        raise Exception("Gap of lacking IMGs (%d) wider than allowed (%d)" % (gap, max_missing))


    # OUTPUT PERIOD => NvO = Nr. of OUT-dekads included
    # # TODO: check Odate1 en Odate 2 voor laatste dag van de dekade en voor laatste dage van de input range
    yyyy, yy, mm, m, tty, ttm, dd, doy, dd_st, dpm, did = parse_date(start_out)
    Odate1 = yyyy + mm + dd_st              # Reset to start of dekad!
    startdekad = dateToDekad(Odate1)
    yyyy, yy, mm, m, tty, ttm, dd, doy, dd_st, dpm, did = parse_date(end_out)
    Odate2 = yyyy + mm + dd_st              #Reset to start of dekad!
    enddekad = dateToDekad(Odate2)

    dekad_lst_out = getDekadLst(startdekad, enddekad)
    NvO = len(dekad_lst_out)                #NvO=Nr. of OUT-dekads (never gaps)
    print ("Nr. of OUT-Dekads   : %s" % NvO)

    Settings_vio1 = -1

    HO=[]
    NvO = 0
    for v in range(NvI):
        dekad = dekad_lst[v]
        dekad_date = HI[v].date
        if ( (dekad_date < int(Odate1)) or (dekad_date > int(Odate2))): continue
        if Settings_vio1 == -1:  Settings_vio1 = v   #CRUCIAL: vio1 = First OUT-IMG counted in terms of IN-IMGs (0 ... NvI-1)

        out_fname = out_file_prefix + dekad + out_file_suffix + '.tif'
        out_qua_fname = '%s%s.tif' % (out_qua_file_prefix, dekad )
        print ("Creating image %s" % out_fname)
        HO.append(HdrInfo_out(out_fname, out_qua_fname, int(dekad_date)))
        NvO += 1

    #OTHER PARAMETERS concerning the TIME SERIES OF IN/OUT-IMGs
    x = hv.Vint + hv.Vslo * hv.Vlo
    y = hv.Vint + hv.Vslo * hv.Vhi
    if ymin  < x or ymin> y:
        raise Exception("ymin should be between [%s %s ]" % (x, y))
    Settings_Ymin = ymin

    #YmaxDip for elimination of local minima  [def=0=skip] =>
    Settings_YmaxDip = ymaxdip
    if Settings_YmaxDip > 0:
        Settings_PreMaxDip = Settings_YmaxDip
    else:
        Settings_PreMaxDip = 1000000
    Settings_PreMaxDif = Settings_PreMaxDip;


    #MaxMiss  NB: This holds for all IN-observations which are initially flagged or with VI < Ymin. More ones may be eliminated in later PRE-stages!
    NviMaxMisPerc = nvimaxmix           #Max. %% of missing values per pixel-profile   [def=75]
    Settings_NviMaxMis = floor(0.5 + NvI * NviMaxMisPerc/100.0)     #Max. Nr. of Missing values (eg. 36 * 75/100 = 27 missing, 9 good)
    if NvG > Settings_NviMaxMis :
        raise Exception("More lacking IN-IMGs (%d) than allowed Missing Values (%d)", NvG, Settings_NviMaxMis)


    #FIXED settings
    Settings_Extrapol        = 1          #Extrapolate 'TAILS' with missing IN-values
    Settings_METflag         = 1          #Still can be changed here to METflag=0 (but not METflag=2)

    Settings_PreMaxGap       = 0          #Keep gaps longer than this nr. of Dekads, reset them to Ymin, presumably SNOW/POLAR WINTER	(def=0       => Skip)
    Settings_PreMaxGapApply  = 0
    Settings_SnowMsk         = 0
    Settings_PreMaxGapMsk    = "".rjust(260).encode('utf-8')    #Optional BYTE SnowMask-IMG: only apply PreMaxGap for pixels with SnowMask-values > 0 (land susceptible to snow/ice).

    Settings_Post            = 0
    Settings_PostOver        = 0
    Settings_PostUnder       = 0
    Settings_PostMax         = 0.

    if( ndvi == 1 ) :                     #1 = NDVI
       Settings_PreMaxTop    = 0.40

       Settings_METsmooth    =   2        #Swets
                                    #Weights for...
       Settings_SwWmax       =   1.5      #- local maxima
       Settings_SwWmin       =   0.005    #- local minima
       Settings_SwWplane     =   1.0      #- planes (same value as 2 neighbours)
       Settings_SwWslope     =   0.5      #- all others in regular profile
       Settings_SwWedge      =   0.5      #- left/right edge point in profile

                                    #SWETS regression parameters (window sizes best ODD - if EVEN, then more past than future IN-periods)
       Settings_SwWinR       = 5          #PERIODs (Dekads) in regression  window: regression parms (A,B) calibrated with these points	[2-36, def=5]
       Settings_SwWinC       = 5          #PERIODs (Dekads) in combination window: regression applied to these points			            [2-36, def=5, min=SwWinR]
       Settings_SwCI         = 0          #Confidence Interval in %, for outlier correction with CHI2-test [def=95%] - Skip this test with SwCI=0 (or 100).
       Settings_MUinterpol = 1  # First replace the missing values by linear interpolation (0=no, 1=def=yes).
       Settings_MUrmf        = 0
    else:                           #0 =  ALBEDO

       Settings_PreMaxTop    = 1000000    #Skip

       Settings_METsmooth    = 0          #MEAN-filter

       Settings_MUinterpol   = 1          #First replace the missing values by linear interpolation (0=no, 1=def=yes).
       Settings_MUrmf        = 0          #Half Length in Dekads of the sliding window used to apply the RMF (0=skip filtering). NB: No default, mandatory KeyWord.
                                          #Half Length MUrmf = 1 / 2 / ... 5 => Full Length = 3 /5 / ... / 11

       #dummy settings for Swets, not used for ALBEDO:
       Settings_SwWmax       =   1.5      #- local maxima
       Settings_SwWmin       =   0.005    #- local minima
       Settings_SwWplane     =   1.0      #- planes (same value as 2 neighbours)
       Settings_SwWslope     =   0.5      #- all others in regular profile
       Settings_SwWedge      =   0.5      #- left/right edge point in profile

       #SWETS regression parameters (window sizes best ODD - if EVEN, then more past than future IN-periods)
       Settings_SwWinR       = 5          #PERIODs (Dekads) in regression  window: regression parms (A,B) calibrated with these points	[2-36, def=5]
       Settings_SwWinC       = 5          #PERIODs (Dekads) in combination window: regression applied to these points			            [2-36, def=5, min=SwWinR]
       Settings_SwCI         = 0          #Confidence Interval in %, for outlier correction with CHI2-test [def=95%] - Skip this test with SwCI=0 (or 100).

    #INTERNAL FLAGS
    Settings_VIflag   = Settings_Ymin - 100.          #Internal Flag (FLOAT) for bad/lacking IN-values in profile-arrays Yi[], Yh[], ...
                                        #VIflag must be smaller than lowest "good" value Ymin
    Settings_FLAGi = 251                       #Default IN-flag

    #----------------------
    # B9. QUALITY OUT-IMGs: Convert fo_pre to _fullpath in view of fo_varmta (which must contain full paths)
    #----------------------
    # TODO switch QUA on or off???
    # Settings_QUA = 0 if(argc == NclParms + bat)  { goto QUAskip }
    Settings_QUA = 1
    Settings_Ydelt = ydelt
    if ydelt < 0 or ydelt > abs(y - x):
        raise Exception("ydelt should be between [0 %s ]" % abs(y - x))

    Settings = SmoothSettings(Settings_METsmooth,
        Settings_METflag,
        Settings_Extrapol,
        Settings_NviMaxMis,
        Settings_QUA,
        Settings_YmaxDip,
        Settings_Ydelt,
        Settings_D50pre,
        Settings_D50nxt,

        Settings_PreMaxDip,
        Settings_PreMaxDif,
        Settings_PreMaxTop,
        Settings_PreMaxGap,
        Settings_PreMaxGapApply,
        Settings_SnowMsk,


        Settings_MUinterpol,
        Settings_MUrmf,
        Settings_SwWmax,
        Settings_SwWmin,
        Settings_SwWplane,
        Settings_SwWslope,
        Settings_SwWedge,
        Settings_SwCI,
        Settings_SwWinR,
        Settings_SwWinC,
        0,
        0,
        0,
        0,

        Settings_Post,
        Settings_PostMax,
        Settings_PostOver,
        Settings_PostUnder,


        Settings_Ymin,
        Settings_VIflag,
        Settings_FLAGi,
        Settings_vio1,
                              Settings_PreMaxGapMsk)
    return HI, HO, Settings


if __name__ == "__main__":
    main()
