# -*- coding: utf-8 -*-
import argparse
import datetime
import rasterio
import numpy as np
from math import floor
from os.path import join
import os
from ctypes import c_short, CDLL, c_float
import ctypes

from frame_s2.config.FRAME_settings import meteo_settings_dict, lv2_settings
from frame_s2.postproc_helper import tif_update_metadata, validate_inputs, dekadToStartDate, get_days_in_dekad, get_dec_end, get_dec_start, parse_date, nextDekad

##### CONSTANTS ############
MAX_NUMBER_IN_FILES = 108   # NvI : Nr. of IN-IMGs: 108 = 3 years * 36 dekads


class Histruc(ctypes.Structure):
    _fields_ = [
        ("date", ctypes.c_long),
        ("fname", ctypes.c_char_p),
        ("lack", ctypes.c_double)
    ]


class HdrInfo(ctypes.Structure):
    _fields_ = [
        ("Vlo", ctypes.c_double),
        ("Vhi", ctypes.c_double),
        ("Vint", ctypes.c_double),
        ("Vslo", ctypes.c_double)
    ]

so_file='tbp.so'
if not os.path.exists(so_file):
    so_file = 'executables/tbp.so'
lib = CDLL(so_file)
# #settings the global Settings structure, these settings are hard coded in C
lib_Set_settings = lib.Set_settings
lib_Set_settings.argtypes = None
lib_Set_settings.restype = None
#
# #settings the global Settings structure, these settings are hard coded in C
lib_Allocate_mem = lib.Allocate_mem
lib_Allocate_mem.argtypes = None
lib_Allocate_mem.restype = None
#
'''#/*void loop(const short *arr_npp_in, /*n_npp_in*rows*columns on npp input data*/
         const int n_npp_in,              /* number of input images, should be 108*/
         const unsigned char *days_in_dekad_in, /*108 times the days_in_dekad*/
         const int rows,      /*nr of rows in matrix*/
         const int columns,   /*nr of columns in matrix*/
         short *arr_npp_pix, /* all 108 inputs per pixel*/
         short *arr_out,        /* rows*columns on out data*/
         const unsigned char *arr_sos,        /* rows*columns on sos data*/
         const unsigned char *arr_eos         /* rows*columns on eos data*/
#*'''
lib_loop = lib.loop
lib_loop.argtypes = [
    np.ctypeslib.ndpointer(dtype=np.int16, ndim=3, flags='C_CONTIGUOUS'),
    c_short,
    c_short,
    #
    c_short,
    c_short,
    np.ctypeslib.ndpointer(dtype=np.int16, ndim=1, flags='C_CONTIGUOUS'),
    np.ctypeslib.ndpointer(dtype=np.float32, ndim=2, flags='C_CONTIGUOUS'),
    np.ctypeslib.ndpointer(dtype=np.ubyte, ndim=2, flags='C_CONTIGUOUS'),
    np.ctypeslib.ndpointer(dtype=np.ubyte, ndim=2, flags='C_CONTIGUOUS')
    ]
lib_loop.restype = None
#T1frac, covers, teg
lib_Set_parameters = lib.Set_Parameters
lib_Set_parameters.argtypes = [HdrInfo, HdrInfo, np.ctypeslib.ndpointer(dtype=np.float32, ndim=1, flags='C_CONTIGUOUS'),c_float, c_float, c_float, c_short, c_short, c_short, c_short]
lib_Set_parameters.restype = None

CFAC = 22.222  # Converts [gC/m�/day] to [kgDM/ha/day]

#
# lib_Set_OutHeaders = lib.Set_OutHeaders
# lib_Set_OutHeaders.argtypes = [HdrInfo, c_short]
# lib_Set_OutHeaders.restype = None


# class SEASON(ctypes.Structure):
#     #Additional parameters: Inputs & Derived
#     _fields_ = [
#         ("SOS", ctypes.c_short), ("MOS", ctypes.c_short), ("EOS", ctypes.c_short), ("L", ctypes.c_short), ("Keep", ctypes.c_short),
#         ("Ysos", ctypes.c_double), ("Ymos", ctypes.c_double), ("Yeos", ctypes.c_double),("Area", ctypes.c_double)
#         ]

# def Set_Global_C_Vars():
#     #SETTING NDVI PARAMETERS
#     print ('---SETTING NDVI IN PARAMETERS----')
#     Vlo=0;Vhi=250;Vint=-0.08;Vslo=0.004;MaxMisPerc=15;
#     #lib_Set_parameters(int(Vlo), int(Vhi), float(Vint), float(Vslo), float(MaxMisPerc))
#
#     #SETTING OUTPUT HEADERS C-var
#     print('---SETTING OUTPUT PARAMETERS----')
#     num = int(0)
#     for key, setting in pheno_settings['out'].items():
#         # self.HDR_ARRAY[num].Vindex = key
#         hdr = HdrInfo(float(setting['Vlo'] or 0), float(setting['Vhi'] or 255), float(setting["Vint"] or 0), float(setting["Vslo"] or 1))
#         #hdr = HdrInfo(8.1*num, 1.0*num, 3.0*num, 4.0*num)
#         lib_Set_OutHeaders(hdr, key)
#         num += 1
#
#     #SETTING other settings
#     lib_Set_settings()
#     lib_Allocate_mem()
#     return


def parse_arguments():
    parser = argparse.ArgumentParser(prog='TBP.py',description="Create TBP output")

    # parser.add_argument("--in_NPP_SMC_lst", type=str, help="a file with 108 NPP_SMC paths, sorted(increasing in time)!!",
    #                     required=False)
    parser.add_argument("--in_file_prefix", type=str,
                        help="NPP dir + file prefix : e.g: /tmp/L2_SRI_NPP_",
                        required=True)
    parser.add_argument("--out_folder", type=str, help="output tbp folder",
                        required=True)
    parser.add_argument("--conversion_factor", type=float, help="Conversion Factor NPP->TBP     (default=22.222)",
                        default=22.222)
    parser.add_argument("--roi", type=str, help="roi=AFR|SRI|SSB|ALG",
                        required=True)
    parser.add_argument("--central_year", type=int, help="Central/Target Year Yt        [YYYY: 1952-2049]",
                        required=True)
    parser.add_argument("--start_in", type=str, help="Start IN-series YYYYMMDD, if=0(default) it will be calculated from the central_year",
                        default='0')
    parser.add_argument("--end_in", type=str, help="End IN-series YYYYMMDD, if=0(default) it will be calculated from the central_year",
                        default='0')
    parser.add_argument("--pheno_file_prefix", type=str, help="file prefix with or without directory: e.g:/tmp/L2_SRI_PHE_2015_",
                        required=True)
    parser.add_argument("--season", type=int, help="Season nr 1|2",
                        required=True)
    parser.add_argument("--teg", type=float, help="Threshold EverGreen (def=0.75, 0=Skip EG-treatment)",
                        default=0.75)
    parser.add_argument("--FENs2yDt", type=int, help="FENs2yDt=0: Civil year covers dekads 37-72 (January-December). But via FENs2yDt, these limits can be shifted up/downwards. FENs2yDt=3: Civil year covers dekads 40-75 (February-January)     (def=3)",
                        default=3)
    # parser.add_argument("--datatype_out", type=int, help="OUT-DataType [2=Short, 3=Short Scaled, 4=Float]",
    #                     default=3)
    parser.add_argument("--date_out", type=str, help="out_date YYYYMMDD, should be the same as end_in",
                        required=True)
    parser.add_argument("--flag_threshold", type=int, help="FLAG DEVIANT PIXELS if their IN-SERIES. Contains < T1% good values   [def=15%, 0=Skip ]",
                        default=15)
    parser.add_argument("--covers", type=int,
                        help="Covers   < T2% of Season Length   [def=0%=Skip]",
                        default=0)
    parser.add_argument("--mask_file", type=str, help="water=0, flag as -9999",
                        required=True)
    return parser.parse_args()

def Set_Global_C_Vars(nr_of_days_per_dekad_lst, T1frac, covers, teg, TTi1, TTi2, T1, TTo, EGdt):
    #SETTING NDVI PARAMETERS
    print ('---SETTING GLOBAL PARAMETERS----')

    #setting parameters
    tbp_hdr = HdrInfo(float(lv2_settings["TBP_VLO"] or 0), float(lv2_settings["TBP_VHI"] or 32767), float(lv2_settings["TBP_OFFSET"] or 0),
                  float(lv2_settings["TBP_SCALE"] or 1))

    npp_scale = lv2_settings["NPP_SMC_SCALE"]*CFAC #Cfac = 22.222 (default). Converts [gC/m2/day] to [kgDM/ha/day]
    npp_offset = lv2_settings["NPP_SMC_OFFSET"]*CFAC
    npp_hdr = HdrInfo(float(lv2_settings["NPP_SMC_VLO"] or 0), float(lv2_settings["NPP_SMC_VHI"] or 32767), float(npp_offset or 0),
                  float(npp_scale or 1))

    lib_Allocate_mem()
    nrs = np.array(nr_of_days_per_dekad_lst).astype(np.float32)
    lib_Set_parameters(tbp_hdr, npp_hdr, nrs, T1frac/100, covers/100, teg, TTi1, TTi2 , T1, TTo, EGdt)
    lib_Set_settings()

def calc_tbp_by_matrix(in_data_cube, nr_in_files, in_sos, in_eos, start_index):
    #calculation phenology using matrixes

    # initialize C-arrays
    n_in, rows, columns = in_data_cube.shape
    in_data_cube = in_data_cube.astype(np.short)
    result_cube = np.full([rows, columns], -1, np.float32)
    arr_pix = np.empty(nr_in_files, np.short)



    #print ("Starting lib loop")
    lib_loop(in_data_cube, nr_in_files, start_index, rows, columns, arr_pix, result_cube, in_sos, in_eos )

    # #TODO
    # ''' """""""""""""""""""
    # if (S.o1 > 0)
    #     {
    #     for (i= 7;i <= 10;i++) {foM[i] = 0;}
    #     for (i=15;i <= 18;i++) {foM[i] = 0;}
    #     for (i=19;i <= 20;i++) {foM[i] = 0;}
    #     }'''
    return result_cube

def main():
    args = parse_arguments()
    #calc_pheno_by_matrix(np.zeros((2,2,2),np.ubyte))
    #start_pheno(args.in_NPP_SMC_lst, args.out_folder, args.roi, args.year)
    start_tbp(args.in_file_prefix, args.out_folder, args.roi, args.central_year, args.start_in, args.end_in, args.season, args.pheno_file_prefix, args.teg, args.FENs2yDt, args.flag_threshold, args.covers, args.date_out, args.mask_file)

def building_NPP_dct(in_fileprefix, central_year, startdate, enddate):
    #The best situation is that we have 3 years of inputs, so 108 dekads
    # - build the full 108 dekad lst, and keep track if the dekad files are lacking
    start_108_dekad = str(central_year -1)[2:4] + '01'  #first dekad of the 3 years on inputs
    end_108_dekad = str(central_year + 1)[2:4] + '36'   #last dekad of the 3 years on inputs
    dekad_counter = 0
    start_dekad_index = 0   #start index of the valid NPP range, (the actual situation), for the full 3 years of data it will be 0
    end_dekad_index = 0     #end index of the valid NPP range, for the full 3 years of data it will be 108
    NiGOOD = 0  #number of valid NPP files
    npp_dict={} #dicdt with final structure


    dekad = start_108_dekad
    while dekad <= end_108_dekad:

        lack=1 #by default the file is not there
        dekad_date = dekadToStartDate(dekad)


        days_in_dekad = get_days_in_dekad(dekad_date)
        fname =  '%s%s.tif' % (in_fileprefix, dekad)
        if os.path.exists(fname):
            lack = 0
        else:
            print ("Lacking npp data for date : %s" %fname)

        # see if date is between startdate & endate
        if dekad_date < startdate or dekad_date > enddate:
            lack = 1

        #fill up final structure and statistics
        fhandler = None
        if lack == 0 :
            NiGOOD += 1
            if start_dekad_index == 0:
                start_dekad_index = dekad_counter
            end_dekad_index = dekad_counter
            fhandler = rasterio.open(fname, 'r', sharing=False)


        npp_dict[dekad_counter] = {
            "fhandler" : fhandler,
            "dekad" : dekad,
            "lack": lack,
            "fname" : fname,
            "startdate" : dekad_date,
            "day_in_dekad": days_in_dekad
        }
        dekad = nextDekad(dekad)
        dekad_counter += 1
    return npp_dict, NiGOOD, start_dekad_index, end_dekad_index

def start_tbp(in_fileprefix, out_folder, roi, central_year, Idate1, Idate2, season, pheno_file_prefix, teg, EGdt, flag_threshold, covers, date_out, mask_file):
    os.environ['GDAL_CACHEMAX'] = meteo_settings_dict['GDAL_CACHEMAX']
    central_year = int(central_year)
    #check settings season+teg
    if season == 1:
        if teg == 0.0:
            raise Exception("Season 1 should be processes with teg=0,75")
    if season ==  2:
        if teg != 0.0:
            raise Exception("Season 2 should be processed with teg=0")

    #checking dates
    l = str(central_year - 1) + '0101'
    ll = str(central_year + 1) + '1231'
    if Idate1 > '0' : #only if it's not the default, otherwise set this later on
        if Idate1 < l or Idate1 > ll:
            raise Exception("Start of IN-series (%s) must be in range [%s-%s]" %(Idate1, l, ll))
        yyyy, yy, mm, m, tty, ttm, dd, doy, dd_st, dpm, did = parse_date(Idate1)
        Idate1 = yyyy + mm + dd_st #Reset to start of dekad!
        #l = Idate1
    if Idate2 > '0':  #Only if it's not the default, otherwise set this later on
        if Idate2 < l or Idate2 > ll:
            raise Exception("Start of IN-series (%s) must be in range [%s-%s]" %(Idate2, l, ll))
        yyyy, yy, mm, m, tty, ttm, dd, doy, dd_st, dpm, did = parse_date(Idate2)
        Idate2 = yyyy + mm + dd_st #Reset to start of dekad!
    nr_of_days_per_dekad_lst = get_days_in_dekad_lst(l, ll)
    # IN-IMGs - CHECKS:   Scan All 108 Dekads in 3 civil Years with Yt in centre
    #- Normally, check from Idate1 to Idate2, but...
    #  . If Idate1=0 (AUTO): check from start    (dekad 1)
    #  . If Idate2=0 (AUTO): check until the end (dekad 108)
    #- Find/memorize Nominal dates & Nr. of days in each dekad
    #- Define/memorize IMG-names
    #- Check which IMGs are present/lacking
    #  Keep track of TTi1 and TTi2 = first/last existing IMG in series
    #- For existing IMGs: read HDR in hh
    #  Memorize hi=HDR of FIRST existing IMG (later replaced by LAST in-HDR)
    #- Checks
    #  . If hh.date is present, then it must be equal to Nominal Date of dekad
    #  . All IN-IMGs must have same spatial features (Ncol/Nrec, map info)
    #  . All IN-IMGs must spectrally conform to definitions in general HDR hm (see p1): Datatype=2 (Short), Vlo/hi, Vint/slo.
    npp_dct, NiGOOD, TTi1, TTi2 = building_NPP_dct(in_fileprefix, central_year, Idate1, Idate2)

    #Some Conclusions/Tests about the IN-files
    Ni = TTi2 - TTi1 + 1    #Dekads in considered period (max. 108)
    print ("Start/End IN-series   : %s - %s" % (Idate1, Idate2))
    print ("Dekads in IN-series   : %s" % Ni)
    print ("Nr. of In-files found : %s" % NiGOOD)
    print ("Lacking in-files      : %s" % (Ni-NiGOOD))														    # Existing IN-IMGs found, between TTi1 and TTi2

    if Ni<20:
        raise Exception("IN-series should at least cover 20 dekads (now: %s)" % Ni)
    if(Ni - NiGOOD > 0):
        raise Exception("IN-series comprises %s lacking in-files" % (Ni - NiGOOD))


    if not (EGdt >= -18 and EGdt <= 18):
        raise Exception("parameter FENs2yDt is %s, but should be between [-18 .. 18]" % EGdt)

    #teg : Threshold EverGreen pixels
    teg_min  = lv2_settings["NPP_SMC_OFFSET"] + lv2_settings["NPP_SMC_SCALE"] * lv2_settings["NPP_SMC_VLO"]    #Min/Max for EGthr
    teg_max = lv2_settings["NPP_SMC_OFFSET"] + lv2_settings["NPP_SMC_SCALE"] * lv2_settings["NPP_SMC_VHI"]
    if teg < float(teg_min) or teg > float(teg_max):
        raise Exception("teg is %s, but should be between [%s ... %s" % (teg, teg_min, teg_max))
    if teg == 0.0:
        EGdt = 0

    #Special flags
    T1frac = flag_threshold / 100
    T1 = floor(0.5 + Ni*T1frac) #Min. Nr. of GOOD values (eg. 108 * 15 / 100 = 15.7 = 16 good dekads out of 108)

    #out_date
    yyyy, yy, mm, m, tty, ttm, dd, doy, dd_st, dpm, did = parse_date(date_out)
    date_out = yyyy + mm + dd_st    # Reset to start of dekad!
    TTo = 36 + (int(yyyy)-central_year)*36 + int(tty)                    # Sequential nr (1-108) of OUT-IMG

    #handling Pheno datafiles
    sos_file = "%ss%s.tif" % (pheno_file_prefix, season)
    eos_file = "%se%s.tif" % (pheno_file_prefix, season)
    validate_inputs([sos_file, eos_file, npp_dct[TTi1+1]['fname'], mask_file], spectral_validation=False)   #check spatial content from pheno files and npp files(only use the first npp for this)
    validate_inputs([sos_file, eos_file], spatial_validation=False, datatype_should_be='uint8')

    #handling OUT's
    #for AFR, the roi abbreviation is not in the filenames, for the other roi's it is.
    if roi == 'AFR':
        roi_prefix = 'L2_TBP_%s_s%s' % (central_year, season)
    else:
        roi_prefix = 'L2_%s_TBP_%s_s%s' % (roi, central_year, season)


    # Creating profile with sos file
    with rasterio.open(sos_file, 'r', sharing=False) as src_sr:
        sr_profile = src_sr.profile

    #set output
    profile = sr_profile
    profile.update(
        dtype=rasterio.int16,
        count=1,
        compress='lzw',
        BIGTIFF='YES')


    #setting global C-parameters
    Set_Global_C_Vars(nr_of_days_per_dekad_lst, flag_threshold, covers, teg, TTi1 +1, TTi2 +1, T1, TTo, EGdt)


    #open outfiles and put the to the outfile lst:
    out_fname = join(out_folder, roi_prefix + '.tif')
    out_fname_vito = join(out_folder, roi_prefix + '_vito.tif')
    dst = rasterio.open(out_fname, 'w', **profile, sharing=False)
    dst_vito = rasterio.open(out_fname_vito, 'w', **profile, sharing=False)


    #Open pheno files sos, eos
    sos_src = rasterio.open(sos_file, 'r', sharing=False)
    eos_src = rasterio.open(eos_file, 'r', sharing=False)
    msk_src = rasterio.open(mask_file, 'r', sharing=False)

    #START BLOCK-based processing
    for ji, window in dst.block_windows(1):
        print ('ji : %s, %s' % (ji[0], ji[1]))
        sos_array = sos_src.read(1, window=window).astype(np.ubyte)
        eos_array = eos_src.read(1, window=window).astype(np.ubyte)
        msk_array = msk_src.read(1, window=window).astype(np.ubyte)

        #Building DATACUBE
        #loop over npp_dict to filter uses inputs
        in_data_cube = np.full((108, window.height, window.width), 0, np.short)
        data_cube_index=0
        for npp_index, npp_setting in npp_dct.items():
            if npp_setting['lack'] == 1: continue     #skip lacking images
            NPP_file = npp_setting['fname']
            src = npp_setting['fhandler']
            npp = src.read(1, window=window).astype(np.short)
            in_data_cube[npp_index, ...] = npp
            data_cube_index += 1
            npp = None


        # Start processing
        tbp_vito = calc_tbp_by_matrix(in_data_cube, 108, sos_array, eos_array, 1 )
        #tbp = calc_tbp_by_matrix(in_data_cube, NiGOOD, sos_array, eos_array, TTi1+1 )
        data_cube = None
        # writing blocks to files
        tbp_vito = np.floor(0.5 + tbp_vito)
        tbp = np.where(tbp_vito < 0, -9996, tbp_vito.astype(int) )
        tbp = np.where(msk_array == 0, -9999, tbp.astype(int))  #set all water data to -9999

        dst_vito.write(tbp_vito.astype('short'), 1, window=window)
        dst.write(tbp.astype('short'), 1, window=window)


    # closing outfile
    dst.close()
    dst_vito.close()
    sos_src.close()
    eos_src.close()
    msk_src.close()

    #closing all opened NPP files
    for npp_index, npp_setting in npp_dct.items():
        if npp_setting['lack'] == 0:     #skip lacking images
            npp_setting['fhandler'].close()

    #update tif metadata
    scale = lv2_settings['TBP_SCALE'],
    offset = lv2_settings['TBP_OFFSET'],
    description = lv2_settings['TBP_DESCRIPTION'].replace('<season>', str(season)).replace('<startdate>', str(Idate1)).replace('<enddate>', str(Idate2)).replace('<central_year>', str(central_year)).replace('<teg>',str(teg)),
    units = lv2_settings['TBP_UNITS'],
    flags = lv2_settings['TBP_FLAGS']
    nodata = lv2_settings['TBP_NODATA']
    processing_info = 'Threshold EverGreen=%s;EGdt=%s;flag_threshold=%s;covers=%s' % (teg, EGdt, flag_threshold, covers)
    vito_flags = '-1=water,-2=missing PHENO,-3=no season 2,-4=insufficient data,-5=before season'
    out_name_vito = tif_update_metadata(out_fname_vito, scale=(scale or 1), offset=(offset or 0), description=description,
                                        units=units, flags=vito_flags, nodata=nodata, processing_info=processing_info)
    out_name = tif_update_metadata(out_fname, scale=(scale or 1), offset=(offset or 0), description=description,
                                   units=units, flags=flags, nodata=nodata, processing_info=processing_info)
    print("Finished successfully with writing %s" % out_name)

#   get_days_in_dekad_lst('20180102','20180515')
def get_days_in_dekad_lst(period_startdate, period_enddate):
    period_enddate=datetime.datetime.strptime(period_enddate, "%Y%m%d")
    start_of_first_dekad = get_dec_start(period_startdate)
    date = datetime.datetime.strptime(start_of_first_dekad, "%Y%m%d")
    result=[]
    while date < period_enddate:
        yyyymmdd = date.strftime("%Y%m%d")
        days = get_days_in_dekad(yyyymmdd)
        dek_end_date = get_dec_end(yyyymmdd)
        result.append(days)
        #set date to first date of next dekad
        date = datetime.datetime.strptime(dek_end_date,"%Y%m%d") + datetime.timedelta(days=1)
    return result


if __name__ == "__main__":
    main()